body {
    font-family: "Droid Serif", serif !important;
    font-size: .9rem;
    line-height: 1.6;
}

p {
    font-size: 1em;
}

div, h1, h2, h3, h4, h5, h6 {
    font-family: "Droid Serif", serif !important;
}

h1 {
    font-weight: 550 !important;
    font-size: 2em;
}

h2 {
    font-weight: 550 !important;
    font-size: 1.75em;
}

h3 {
    font-weight: 550 !important;
    font-size: 1.5em;
}

h3 {
    font-weight: 550 !important;
    font-size: 1.25em;
}

h4 {
    font-weight: 550 !important;
    font-size: 1.1em;
}

figure figcaption {
    font-family: "Droid Serif", serif !important;
}

figure figcaption p {
    font-size: .8em !important;
}

div.citation {
    font-size: .8rem;
    line-height: 1.5;
}

div.citation p {
    font-size: .8rem;
}

.bd-sidebar-primary {
    font-size: .9em;
}

.navbar-nav li {
    line-height: 1.2em;
}

.page-toc .nav-link {
    font-size: .9em;
}

.page-toc {
    line-height: 1.4;
}

.admonition p.admonition-title ~ *,
div.admonition p.admonition-title ~ * {
  margin-left: 0rem;
  margin-right: 0rem;
}

.MathJax {
    font-size: 1.2em !important;
}

.version-switcher__container {
    text-align: center;
}

.bd-sidebar-primary .navbar-icon-links {
  justify-content: center !important;
}

